'''
https://www.reuters.com/finance/stocks/overview/VUKE.F
'''
import logging
from pricedb import SecuritySymbol, PriceModel


class Reuters:
    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def download(self, symbol: SecuritySymbol, currency: str) -> PriceModel:
        import urllib.request

        url = self.get_security_url(symbol)
        # download
        with urllib.request.urlopen(url) as response:
            html = response.read()
        if not html:
            return None

        price = self.parse_price(html)
        price.symbol = symbol

        return price

    def get_security_url(self, security: SecuritySymbol) -> str:
        base = "https://www.reuters.com/finance/stocks/overview/"

        exchange_dict = {
            'FWB': 'F'
        }

        namespace = exchange_dict[security.namespace]

        return f"{base}{security.mnemonic}.{namespace}"
    
    def parse_price(self, html: str) -> PriceModel:
        from decimal import Decimal, InvalidOperation
        from bs4 import BeautifulSoup

        result = PriceModel()
        soup = BeautifulSoup(html, 'html.parser')

        price_el = (soup.find(id='headerQuoteContainer')
            .find('div', class_='sectionQuoteDetail')
            .find_all('span'))

        # price value
        price_text = price_el[1]
        value = price_text.get_text().strip()
        try:
            result.value = Decimal(value)
        except InvalidOperation:
            message = f"Could not parse price value {value}"
            print(message)
            self.logger.error(message)
            return None

        # currency
        cur_text = price_el[2].get_text().strip()
        result.currency = cur_text

        # date / time
        date_str = price_el[3].get_text().strip()
        result.datum = self.parse_date(date_str)
        
        return result

    def parse_date(self, date_str: str):
        ''' parse date/time '''
        from dateutil.parser import parse
        from dateutil import tz
        from pydatum import Datum

        # Can be "19 Jul 2019" or "6:06am EDT".

        if "EDT" in date_str:
            # the format is "6:06am EDT"
            from_zone = tz.gettz('US/Eastern')
            #est = tz.gettz('EST')
            to_zone = tz.tzlocal()
            parsed_date = parse(date_str)
            date_val = parsed_date.replace(tzinfo=from_zone).astimezone(tz=to_zone)
        else:
            # the date format is "19 Jul 2019"
            #date_val = datetime.strptime(date_str, "%d %b %Y")
            date_val = parse(date_str)

        datum = Datum()
        result = datum.from_datetime(date_val)
        return result


""" The main module """
from decimal import Decimal
from typing import List
import logging
from enum import Enum, auto

from pricedb import PriceModel


class DownloadSources(Enum):
    """ Available sources for price download """
    alphavantage = auto(),
    fixerio = auto(),
    boerse_frankfurt = auto(),
    morningstar = auto(),
    morningstar_de = auto(),
    reuters = auto(),
    vanguard_au = auto()


class Quote:
    """ The main application object. 
    Select the downloading source first q.set_source(<name>)
    then call q.fetch("ASX", "VAS").
    The list of sources is available through q.sources property.
    """
    symbol = None
    exchange = None
    source = None
    _currency = None

    """ The main application object """
    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def fetch(self, exchange: str, symbols: List[str]):
        """ The main method to fetch prices.
        exchange = the name of the exchange, commodity *namespace* in GnuCash;
        symbols = the list of symbols (just the company ticker!) to fetch;
        """
        assert isinstance(exchange, str)
        #assert isinstance(symbols, List[str])
        #assert symbols

        if not symbols:
            raise ValueError("The symbols are missing.")
        
        # fetch the prices using the given module.
        result = []
        for symbol in symbols:
            price = self.__download(exchange, symbol, self._currency, self.source)
            if price:
                result.append(price)

        return result

    def currency(self, source: str, destination: str) -> PriceModel:
        """ Fetches the currency exchange rate of the 1st currency in 2nd.
        i.e. currency("AUD", "EUR") will return the rate for AUD in Euros. (0.71)
        """
        source = source.upper()
        destination = destination.upper()

        # the default currency rate provider
        default_provider = "Fixerio"

        result = self.__download("CURRENCY", source, destination, default_provider)

        return result

    @property
    def sources(self) -> List[str]:
        """ Available sources for prices """
        result = []
        for source in DownloadSources:
            result.append(source.name)
        return result

    def set_currency(self, currency: str):
        """ Sets the output currency """
        assert currency

        self._currency = currency.upper()

    def set_source(self, source: str):
        """ Set the download source to use to fetch the prices. """
        self.source = source.lower()

    def __download(self, exchange: str, symbol: str, currency: str = None, source: str = None):
        """ Download single latest price """
        from pricedb import SecuritySymbol
        from .alphavantage import AlphaVantageDownloader
        from .morningstar import MorningstarDownloader
        from .morningstar_de import MorningstarDeDownloader
        from .vanguard_au import VanguardAuDownloader
        from .boerse_frankfurt import FwbDownloader
        from .fixerio import Fixerio
        from .reuters import Reuters

        assert source is not None
        assert isinstance(source, str)

        if exchange:
            exchange = exchange.upper()
        symbol = symbol.upper()
        if currency:
            currency = currency.upper()
        source = source.lower()
        actor = None
        price = None

        security_symbol = SecuritySymbol("", "")
        security_symbol.parse(symbol)
        if not security_symbol.namespace:
            security_symbol.namespace = exchange

        # todo: read import modules dynamically?

        if source == DownloadSources.morningstar.name:
            actor = MorningstarDownloader()
        elif source == DownloadSources.morningstar_de.name:
            actor = MorningstarDeDownloader()
        elif source == DownloadSources.vanguard_au.name:
            actor = VanguardAuDownloader()
        elif source == DownloadSources.alphavantage.name:
            actor = AlphaVantageDownloader()
        elif source == DownloadSources.fixerio.name:
            actor = Fixerio()
        elif source == DownloadSources.boerse_frankfurt.name:
            actor = FwbDownloader()
        elif source == DownloadSources.reuters.name:
            actor = Reuters()
        else:
            raise ValueError("No source specified for price download.")

        if actor:
            actor.logger = self.logger
            try:
                price = actor.download(security_symbol, currency)
            except AttributeError as e:
                self.logger.error(str(e))

        return price



"""
Morningstar.de price downloader
"""
import logging
from datetime import datetime
from decimal import Decimal, InvalidOperation
from pydatum import Datum

from pricedb.model import PriceModel, SecuritySymbol


class MorningstarDeDownloader:
    """ Fetches prices from Morningstar site """
    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def download(self, symbol: SecuritySymbol, currency: str) -> PriceModel:
        """ Download the given symbol """
        import urllib.parse
        import urllib.request

        if not symbol.namespace:
            raise ValueError(f"Namespace not sent for {symbol}")

        codes = {
            "BVME:EMCR": "0P0000YWT1",
            "BVME:DGSE": "0P0001590A",
            "AMS:VUCP": "0P00017M0P",
            "FWB:VUKE": "0P0001BVG4",
            "AMS:VUTY": "0P00017M0O",
            "XETRA:WTES": "0P00015GSJ"
        }
        code = codes[str(symbol)]

        url = f"http://www.morningstar.de/de/etf/snapshot/snapshot.aspx?id={code}"
        self.logger.debug(f"fetching price from {url}")

        with urllib.request.urlopen(url) as response:
            html = response.read()

        if not html:
            return None

        # parse HTML
        price = self.parse_price(html)
        if price:
            price.symbol = symbol
        # compare currency
        if price.currency != currency:
            raise ValueError(f"Requested currency ({currency}) does not match the {symbol} -> {currency}.")

        return price

    def parse_price(self, html: str) -> PriceModel:
        """ parse html to get the price """
        from bs4 import BeautifulSoup

        result = PriceModel()
        soup = BeautifulSoup(html, 'html.parser')

        overview = soup.find(id='overviewQuickstatsDiv')

        # Price
        price_el = overview.find('table').find_all('tr')[1].find_all('td')[2]
        if not price_el:
            logging.debug(f"Received from mstar: {html}")
            raise ValueError("No price info found in returned HTML.")

        value = price_el.get_text().strip()
        
        # Currency
        currency = value[:3]
        result.currency = currency

        value = value[3:].strip()
        # decimal separator
        value = value.replace(',', '.')
        # parse
        try:
            result.value = Decimal(value)
        except InvalidOperation:
            message = f"Could not parse price value {value}"
            print(message)
            self.logger.error(message)
            return None

        # Date
        date_el = overview.find('table').find_all('tr')[1].find_all('td')[0].find('span')
        #date_str = soup.find(id="asOfDate").get_text().strip()
        date_str = date_el.get_text().strip()
        date_val = datetime.strptime(date_str, "%d.%m.%Y")
        result.datum = Datum()
        result.datum.from_datetime(date_val)

        return result



'''
Price downloader for Boerse Frankfurt (FWB)

The site is using Server-Side Events to send price updates.
See https://pypi.org/project/sseclient/

The new boerse-frankfurt.de
https://www.boerse-frankfurt.de/etp/IE00B810Q511
https://api.boerse-frankfurt.de/data/price_information?isin=IE00B810Q511&mic=XETR
https://api.boerse-frankfurt.de/data/quote_box?isin=IE00BKX55Q28&mic=XFRA
'''
import logging
from pricedb import SecuritySymbol, PriceModel


class FwbDownloader:
    ''' FWB '''
    def __init__(self):
        self.namespace = "FWB"
        self.currency = "EUR"
        self.logger = logging.getLogger(__name__)

    def download(self, symbol: SecuritySymbol, currency: str) -> PriceModel:
        ''' download the price '''
        import urllib.parse
        import urllib.request
        from sseclient import SSEClient

        if not symbol.namespace:
            raise ValueError(f"Namespace not sent for {symbol}")

        self.logger.debug(f"fetching price from FWB.")

        url = self.get_security_url(symbol)

        # download
        messages = SSEClient(url)

        if not messages:
            return None

        data = None
        for msg in messages:
            if msg.data:
                data = msg.data
                if "EMPTY" not in data:
                    break

        # parse
        price = self.parse_price(data)

        price.symbol = symbol

        return price

    def get_security_url(self, security: SecuritySymbol) -> str:
        ''' Mapping the security to the price URL '''
        if security.namespace != self.namespace:
            raise ValueError("Wrong exchange requested!")

        sec_codes = {
            "VGOV": "IE00B42WWV65",
            "VMID": "IE00BKX55Q28",
            "VUKE": "IE00B810Q511"
        }
        isin = sec_codes[security.mnemonic]

        #url = f"https://www.boerse-frankfurt.de/etp/{isin}"
        url = f"https://api.boerse-frankfurt.de/data/quote_box?isin={isin}&mic=XFRA"

        return url

    def parse_price(self, json_str: str) -> PriceModel:
        ''' Get the price from HTML '''
        import json
        from decimal import Decimal
        from datetime import datetime
        from pydatum import Datum

        content = json.loads(json_str)

        result = PriceModel()
        price_str = str(content['lastPrice'])
        result.value = Decimal(price_str)
        result.currency = self.currency

        date_str = content['timestampLastPrice']
        date_val = datetime.fromisoformat(date_str)
        result.datum = Datum()
        result.datum.from_datetime(date_val)

        return result
